"""
Initializes values for a particle
* `model`: model containing a likelihood function with data and priors
* `de`: differential evolution object
* `p`: a particle
* `n_iter`: the number of iterations
"""
function init_particle!(model, de, p, n_iter)
    N = de.discard_burnin ? n_iter - de.burnin : n_iter
    p.samples = typeof(p.samples)(undef, N, length(p.Θ))
    p.accept = fill(false, N)
    p.weight = priorlike(model, p) + model.model(p.Θ)
    p.lp = fill(0.0, N)
    return nothing
end

# function priorlike(model, p::Particle{Array{T,1}}) where {T <: Real}
#     return sum(logpdf.(model.priors, p.Θ))
# end

function priorlike(model, p)
    LL = 0.0
    for (pr,θ) in zip(model.priors, p.Θ)
        LL += get_LL(pr[1], θ)
    end
    return LL
end

# handles arbitrary distribution types. Not flexible enough otherwise
get_LL(d::MultivariateDistribution, x::Vector) = logpdf(d, x)
get_LL(d::UnivariateDistribution, x::Real) = logpdf(d, x)
get_LL(d, x) = loglikelihood(d, x)

"""
Metropolis-Hastings proposal selection
Note: assumes weights are posterior log likelihoods
* `proposal`: weight of proposal e.g. posterior log likelihood
* `current`: weight of current value e.g. posterior log likelihood
"""
function accept(proposal, current)
    p = min(1.0, exp(proposal - current))
    rand() <= p ? (return true) : (return false)
end

"""
Checks whether parameter is within lower and upper bounds
* `b`: boundary (lowerbound,upperbound)
* `θ`: a parameter value
"""
in_bounds(b, θ::Real) = θ >= b[1] && θ <= b[2]
in_bounds(b, θ::Array{<:Real,N}) where {N} = all(x -> in_bounds(b, x), θ)

function in_bounds(de::DE, proposal)
    for (b,θ) in zip(de.bounds, proposal.Θ)
        !in_bounds(b, θ) ? (return false) : nothing
    end
    return true
end

function compute_posterior!(de, model, proposal)
    if in_bounds(de, proposal)
        proposal.weight = priorlike(model, proposal) + model.model(proposal.Θ)
    else
        proposal.weight = -Inf
    end
    return nothing
end

"""
Returns parameters names.
* `model`: model containing a likelihood function with data and priors
* `p`: a particle
"""
function get_names(model, p)
    N = size.(p.Θ)
    n_parms = length.(p.Θ)
    parm_names = fill("", sum(n_parms))
    cnt = 0
    for (k,n) in zip(model.names, N)
        if isempty(n)
            cnt += 1
            parm_names[cnt] = string(k)
        else
            for i in CartesianIndices(n)
                cnt += 1
                parm_names[cnt] = string(k, "[", join([i.I...], ","), "]")
            end
        end
    end
    push!(parm_names, "acceptance", "lp")
    return parm_names
end

"""
Store samples after burnin period
Selects between mutation and crossover step with probability β
* `de`: differential evolution object
* `groups`: groups of particles
"""
function store_samples!(de, groups)
    (de.iter <= de.burnin && de.discard_burnin) ? (return) : nothing
    i = de.discard_burnin ? de.iter - de.burnin : de.iter
    for group in groups
        for p in group
            add_sample!(p, i)
        end
    end
    return nothing
end

function add_sample!(p::Particle{T}, i) where {T <: Real}
    p.samples[i,:] = p.Θ'
end

function add_sample!(p, i)
    p.samples[i,:] = p.Θ
end

function sample_prior(priors)
    samples = [rand(p...) for p in priors]
    t = findtype(samples)
    return sample_prior(samples, t)
end

sample_prior(samples, T) = T[s for s in samples]

findtype(p) = Union{unique(typeof.(p))...}

"""
Update particle based on Metropolis-Hastings rule.
* `de`: differential evolution object
* `current`: current particle
* `proposal`: proposal particle
"""
function update_particle!(de, current, proposal)
    @unpack iter,burnin = de
    i = de.discard_burnin ? iter - burnin : iter
    accepted = accept(proposal.weight, current.weight)
    if accepted
        current.Θ = proposal.Θ
        current.weight = proposal.weight
    end
    if i > 0
        current.accept[i] = accepted
        current.lp[i] = current.weight
     end
    return nothing
end


# Type-stable arithmatic operations for Union{Array{Float64,1},Float64} types (which return Any otherwise)
import Base: +, - ,*

function +(x::Particle, y::Particle)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] + y.Θ[i]
    end
    return Particle(Θ=z)
end

+(x::Real, y::Particle) = +(y, x)

function +(x::Particle, y::Real)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .+ y
    end
    return Particle(Θ=z)
end

function +(x::Particle, d::Distribution)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .+′ draw(d, x.Θ[i])
    end
    return Particle(Θ=z)
end

function draw(d, v::Float64)
    return rand(d)
end

function draw(d, v)
    return rand(d, size(v))
end

function *(x::Particle, y::Particle)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .*′ y.Θ[i]
    end
    return Particle(Θ=z)
end

*(x::Real, y::Particle) = *(y, x)

function *(x::Particle, y::Real)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .*′ y
    end
    return Particle(Θ=z)
end

*(x::Array{<:Real,1}, y::Particle) = *(y, x)

function *(x::Particle, y::Array{<:Real,1})
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .* y[i]
    end
    return Particle(Θ=z)
end

function -(x::Particle, y::Particle)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] - y.Θ[i]
    end
    return Particle(Θ=z)
end

-(x::Real, y::Particle) = -(y, x)

function -(x::Particle, y::Real)
    N = length(x.Θ)
    z = similar(x.Θ)
    for i in 1:N
        z[i] = x.Θ[i] .- y
    end
    return Particle(Θ=z)
end

# arithmetic methods for hanlding discrete parameters
*′(x, y) = x * y
*′(x::Int64, y::Float64) = Int(round(x * y))
*′(x::Float64, y::Int64) = Int(round(x * y))
*′(x::Array{Int64,N}, y::Float64) where {N} = @. Int(round(x * y))
*′(x::Float64, y::Array{Int64,N}) where {N} = @. Int(round(x * y))
+′(x, y) = x + y
+′(x::Int64, y::Float64) = Int(round(x + y))
+′(x::Float64, y::Int64) = Int(round(x + y))
+′(x::Array{Int64,N}, y::Float64) where {N} = @. Int(round(x + y))
+′(x::Float64, y::Array{Int64,N}) where {N} = @. Int(round(x + y))